

"""
Collecting handled Exceptions used by CheckQC.
"""


class CheckQCException(Exception):
    pass


class ConfigurationError(CheckQCException):
    pass


class InstrumentTypeUnknown(ConfigurationError):
    pass


class RunModeUnknown(ConfigurationError):
    pass


class ReagentVersionUnknown(ConfigurationError):
    pass


class RunInfoXMLNotFound(CheckQCException):
    pass


class DemuxSummaryNotFound(CheckQCException):
    pass


class RunParametersNotFound(CheckQCException):
    pass


class StatsJsonNotFound(CheckQCException):
    pass


class QCHandlerNotFound(CheckQCException):
    pass


class ConfigurationError(CheckQCException):
    pass


class SamplesheetNotFound(CheckQCException):
    pass


class ConfigEntryMissing(ConfigurationError):
    pass


class RunfolderNotFoundError(CheckQCException):
    pass




import logging

from checkQC.exceptions import ConfigurationError

log = logging.getLogger(__name__)


class QCHandlerReport(object):
    """
    Base class of objects which contain reports from a QCHandler
    """

    def __init__(self, msg, ordering=1, data=None):
        """
        Instantiate a new QCHandlerReport

        :param msg: Message to send
        :param ordering: A number indicating the ordering of this object in relation to other objects of the same type
                        can be used e.g. to order reports by lane
        :param data: Any additional data (this can e.g. be used to dump additional data into a json object), should be
                     a dict
        """
        self.message = msg
        self.ordering = ordering
        self.data = data

    def type(self):
        """
        Should be implemented by the subclass.

        :returns: String with the type of the report, e.g. "error" or "warning"
        """
        raise NotImplementedError("Subclass must implement this method")

    def __repr__(self):
        """
        Providing string representation of this class

        :returns: Class string representation
        """
        return self.__str__()

    def as_dict(self):
        """
        Dump the class as a dictionary

        :returns: A dict of representing this QCHandlerReport
        """
        return {'type': self.type(), 'message': self.message, 'data': self.data}


class QCErrorFatal(QCHandlerReport):
    """
    Class representing a fatal QC error from a handler, i.e. a value which should in the end
    yield a non-zero exit status from the program.
    """

    def __init__(self, msg, ordering=1, data=None):
        super().__init__(msg, ordering, data)

    def __str__(self):
        return "Fatal QC error: {}".format(self.message)

    def type(self):
        return "error"


class QCErrorWarning(QCHandlerReport):
    """
    Class representing a QC warning from a handler, i.e. a value is interesting to note,
    but which should still yield a zero exit status from the program.
    """
    def __init__(self, msg, ordering=1, data=None):
        super().__init__(msg, ordering, data)

    def __str__(self):
        return "QC warning: {}".format(self.message)

    def type(self):
        return "warning"


class Subscriber(object):
    """
    Subscriber defines the behaviour necessary to subscribe to data from a Parser. The implementing subclass
    has to implement the `collect` method. This method can decide which objects sent to the Subscriber that
    are of interest to that particular Subscriber, and what should be done with those values.
    """

    def __init__(self):
        self.subscriber = self.subscribe()
        next(self.subscriber)

    def subscribe(self):
        """
        This method picks up data from the parser to which the Subscriber is listening.

        :returns: None
        """
        while True:
            value = yield
            self.collect(value)

    def collect(self, signal):
        """
        The implementing subclass should provide this method. It is up to instance receiving the data to decide
        how to handle it. Below is an example of how to handle a tuple with a key-value pair.

        .. code-block :: python

            class MySubscriber(Subscriber):

                def __init__(self):
                    self.results = []

                def collect(self, signal):
                    key, value = signal
                    if key == "my_key":
                        self.results.append(value)


        :returns: None
        """
        raise NotImplementedError("Implementing class must provide this method.")

    def send(self, value):
        """
        Will send the specified value to the subscriber

        :param value: Value to send to subscriber
        :returns: None
        """
        self.subscriber.send(value)


class QCHandler(Subscriber):
    """
    The QCHandler is one of the fundamental classes of CheckQC. It is the base class for the the concrete
    implementations of QCHandlers which actually check the the quality criteria of a runfolder.
    For examples of how to implement a QCHandler it is easiest to look at the implementations available in
    the `checkQC.handlers` module
    """

    UNKNOWN = 'unknown'
    ERROR = 'error'
    WARNING = 'warning'

    def __init__(self, qc_config):
        """
        Create a QCHandler instance

        :param qc_config: dict containing the configuration for the QCHandler. Should have keys 'error' and 'warning'
        """
        super().__init__()
        self._exit_status = 0
        self.qc_config = qc_config

    def custom_configuration_validation(self):
        """
        Override this method in subclass to provide additional configuration behaviour.

        :raises: ConfigurationError if there is a problem with the configuration
        """
        pass

    def validate_configuration(self):
        """
        This method will validate the configuration which has been passed to the QCHandler. This should be called
        by the class making use of this instance. It will not be called automatically e.g. at object creation.

        :returns: None
        :raises: ConfigurationError if there is a problem with the configuration
        """
        try:
            self.qc_config[self.ERROR]
            self.qc_config[self.WARNING]
        except KeyError as e:
            raise ConfigurationError("Configuration expects key: {}. Perhaps it is missing?".format(e.args[0]))
        self.custom_configuration_validation()

    def error(self):
        """
        The value associated with a QC error

        :returns: The configuration value for an error
        """
        return self.qc_config[self.ERROR]

    def warning(self):
        """
        The value associated with a QC warning

        :returns: The configuration value for an warning
        """
        return self.qc_config[self.WARNING]

    def exit_status(self):
        """
        The exit status of the handler.

        :returns: 0 if the qc criteria have not encountered a fatal qc error, else 1.
        """
        return self._exit_status

    def parser(self):
        """
        The class of the Parser (or a list of parsers) which this QCHandler will get its data from.
        E.g.

        .. code-block :: python

            def parser(self):
                return InteropParser

        Note that there should be no parenthesis after the class.

        :returns: The Parser implementation needed by this QCHandler
        """
        raise NotImplementedError("A handler needs to return the class of the parser it needs!")

    def check_qc(self):
        """
        The check_qc method provides the core behaviour of the QCHandler. It should check the values provided
        to it and yield instances of QCHandlerReport (or continue, if there was nothing to report)

        :returns: An instance of QCHandlerReport
        """
        raise NotImplementedError("A handler must provide its own QC checking behaviour by implementing "
                                  "the `check_qc` method.")

    def report(self):
        """
        Check the quality criteria as specified in `check_qc` and gather all reports. Will set the objects
        `exit_status` in accordance with what types of reports are found.

        :returns: A sorted list of errors and warnings found when evaluating the qc criteria.
        """
        errors_and_warnings = self.check_qc()
        sorted_errors_and_warnings = sorted(errors_and_warnings, key=lambda x: x.ordering)

        for element in sorted_errors_and_warnings:
            if isinstance(element, QCErrorFatal):
                self._exit_status = 1
                log.error(element)
            else:
                log.warning(element)

        return sorted_errors_and_warnings





from collections import defaultdict
import logging

from checkQC.handlers.qc_handler_factory import QCHandlerFactory
from checkQC.exceptions import ConfigurationError

log = logging.getLogger(__name__)


class QCEngine(object):
    """
    The QCEngine will provide a method to apply all specified handler on the specified runfolder.

    Internally it will run a number of methods which will do the following:
     - create all handlers specified in the handler config
     - validate all the configs provided, so that all necessary values are preset
     - initiate the parsers based on which parsers are found in the handlers
     - connect the handlers and parsers so that each parser gets the correct
       subscribers
     - run the parsers, i.e. pick up data and pass it to the handlers
     - compile all reports from the handlers

    The QCEngine has a `exit_status` field which can be checked after calling the `run` method,
    to determine if all handlers were successful or not (zero indicates success, 1 indicates failure)
    """

    def __init__(self, runfolder, parser_configurations, handler_config, qc_handler_factory=None):
        """
        Create a instance of QCEngine

        :param runfolder: the path to the runfolder which should be analyzed
        :param parser_configurations: dict containing configurations for the parsers
        :param handler_config: a dict which configurations for the handlers
        :param qc_handler_factory: A QCHandlerFactory, if None default QCHandlerFactory will be used
        """
        self.runfolder = runfolder
        self.parser_configurations = parser_configurations
        self.handlers_config = handler_config
        self._handlers = []
        self._parsers_and_handlers = defaultdict(list)
        self.exit_status = 0
        if qc_handler_factory:
            self._qc_handler_factory = qc_handler_factory
        else:
            self._qc_handler_factory = QCHandlerFactory()

    def run(self):
        """
        Run the specified parsers and handlers and compile their reports. Will set the `exit_status` depending
         on if there were any errors or not.

        :return: a dict representing the reports gathers.
        """
        try:
            self._create_handlers()
            self._validate_configurations()
            self._initiate_parsers()
            self._subscribe_handlers_to_parsers()
            self._run_parsers()
            reports = self._compile_reports()
            return reports
        except ConfigurationError:
            self.exit_status = 1

    def _create_handlers(self):
        for clazz_config in self.handlers_config:
            self._handlers.append(self._qc_handler_factory.
                                  create_subclass_instance(clazz_config["name"], clazz_config))

    def _validate_configurations(self):
        try:
            for handler in self._handlers:
                handler.validate_configuration()
            return True
        except ConfigurationError as e:
            log.error("Error in configuration found for handler: {}. {}".format(type(handler).__name__, e))
            raise e

    def _initiate_parsers(self):
        for handler in self._handlers:
            if isinstance(handler.parser(), list):
                parsers = handler.parser()
            else:
                parsers = [handler.parser()]

            for parser_factory in parsers:
                parser_instance = parser_factory(runfolder=self.runfolder,
                                                 parser_configurations=self.parser_configurations)
                self._parsers_and_handlers[parser_instance].append(handler)

    def _subscribe_handlers_to_parsers(self):
        for parser, handlers in self._parsers_and_handlers.items():
            parser.add_subscribers(handlers)

    def _run_parsers(self):
        for parser in self._parsers_and_handlers.keys():
            parser.run()

    def _compile_reports(self):
        reports = {"exit_status": 0}
        for handler in self._handlers:
            handler_report = handler.report()
            if handler_report:
                reports[type(handler).__name__] = list(map(lambda x: x.as_dict(), handler_report))
            if handler.exit_status() != 0:
                self.exit_status = 1
                reports["exit_status"] = 1
        return reports




import logging

log = logging.getLogger(__name__)


class RunTypeSummarizer(object):
    """
    RunTypeSummarizer contains methods used to summarize information about a run (mostly in order
    to be able to report this back to the user in one for or another).
    """

    @staticmethod
    def summarize(instrument_and_reagent_version, read_lengths, handler_config):
        """
        Utility method to parse log and summarize the information about a particular run.
        This is used to tell the user what information CheckQC has gathered about a run, and
        what configuration was used.

        :param instrument_and_reagent_version: str with instrument and reagent version
        :param read_lengths: str with read length
        :param handler_config: dict with handler configuration
        :returns: dict with information on the handlers used
        """

        summary = {}

        log.info("Run summary")
        log.info("-----------")
        log.info("Instrument and reagent version: {}".format(instrument_and_reagent_version))
        summary["instrument_and_reagent_type"] = instrument_and_reagent_version
        log.info("Read length: {}".format(read_lengths))
        summary["read_length"] = read_lengths
        log.info("Enabled handlers and their config values were: ")
        summary["handlers"] = []
        for handler in handler_config:
            log.info("\t{}".format(handler))
            summary["handlers"].append(handler)

        return summary




import sys
import json
import logging
from pathlib import Path

import click

from checkQC.qc_engine import QCEngine
from checkQC.config import ConfigFactory
from checkQC.run_type_recognizer import RunTypeRecognizer
from checkQC.run_type_summarizer import RunTypeSummarizer
from checkQC.exceptions import CheckQCException, RunfolderNotFoundError
from checkQC import __version__ as checkqc_version


logging.basicConfig(level=logging.INFO,
                    format='%(levelname)-8s %(message)s')

console_log_handler = logging.StreamHandler()
log = logging.getLogger(__name__)


@click.command("checkqc")
@click.option("--config", help="Path to the checkQC configuration file", type=click.Path())
@click.option('--json', is_flag=True, default=False, help="Print the results of the run as json to stdout")
@click.option('--downgrade-errors', type=str, multiple=True, help="Downgrade errors to warnings for a specific handler, can be used multiple times")
@click.option('--use-closest-read-length', is_flag=True, default=False, help="Use the closest read length if the read length used isn't specified in the config")
@click.version_option(checkqc_version)
@click.argument('runfolder', type=click.Path())
def start(config, json, downgrade_errors, use_closest_read_length, runfolder):
    """
    checkQC is a command line utility designed to quickly gather and assess quality control metrics from an
    Illumina sequencing run. It is highly customizable and which quality controls modules should be run
    for a particular run type should be specified in the provided configuration file.
    """
    # -----------------------------------
    # This is the application entry point
    # -----------------------------------
    app = App(runfolder, config, json, downgrade_errors, use_closest_read_length)
    app.run()
    sys.exit(app.exit_status)


class App(object):
    """
    This is the main application object for CheckQC.
    """

    def __init__(self, runfolder, config_file=None, json_mode=False,
                 downgrade_errors_for=(), use_closest_read_length=False):
        self._runfolder = runfolder
        self._config_file = config_file
        self._json_mode = json_mode
        self._downgrade_errors_for = downgrade_errors_for
        self._use_closest_read_length = use_closest_read_length
        self.exit_status = 0

    def configure_and_run(self):
        """
        Configures and runs the application. It will set the exit status of the object in accordance with if any
        fatal qc errors were found or not.

        :returns: The reports of the application as a dict
        """
        config = ConfigFactory.from_config_path(self._config_file)
        parser_configurations = config.get("parser_configurations", None)

        if not Path(self._runfolder).is_dir():
            raise RunfolderNotFoundError("Could not find runfolder: {}. Are you "
                                         "sure the path is correct?".format(self._runfolder))

        run_type_recognizer = RunTypeRecognizer(runfolder=self._runfolder)
        instrument_and_reagent_version = run_type_recognizer.instrument_and_reagent_version()

        # TODO For now assume symmetric read lengths
        both_read_lengths = run_type_recognizer.read_length()
        read_length = int(both_read_lengths.split("-")[0])
        handler_config = config.get_handler_configs(instrument_and_reagent_version, read_length,
                                                    self._downgrade_errors_for, self._use_closest_read_length)

        run_type_summary = RunTypeSummarizer.summarize(instrument_and_reagent_version, both_read_lengths, handler_config)

        qc_engine = QCEngine(runfolder=self._runfolder,
                             parser_configurations=parser_configurations,
                             handler_config=handler_config)
        reports = qc_engine.run()
        reports["run_summary"] = run_type_summary
        self.exit_status = qc_engine.exit_status
        return reports

    def run(self):
        """
        This method will run CheckQC as it is intended to run as a commandline application, it will log to the
        stderr and write data to stdout.

        :returns: the exit status of the run (0 for success, else not 0)
        """
        log.info("------------------------")
        log.info("Starting checkQC ({})".format(checkqc_version))
        log.info("------------------------")
        log.info("Runfolder is: {}".format(self._runfolder))
        try:
            reports = self.configure_and_run()
            if self.exit_status == 0:
                log.info("Finished without finding any fatal qc errors.")
            else:
                log.error("Finished with fatal qc errors and will exit with non-zero exit status.")

            if self._json_mode:
                print(json.dumps(reports))

        except CheckQCException as e:
            log.error(e)
            self.exit_status

        return self.exit_status

if __name__ == '__main__':
    start()




from pathlib import Path

from checkQC.runfolder_reader import RunfolderReader
from checkQC.parsers.parser import Parser
from checkQC.exceptions import ConfigurationError, DemuxSummaryNotFound


class DemuxSummaryParser(Parser):
    """
    The DemuxSummaryParser will read information from the DemuxSummaryF1L<Lane number>.txt files.
    At the moment it fetches the information about 'Most Popular Unknown Index Sequences'.

    It will send the information lane wise as a tuple on the format:
     ("index_counts", {"lane": <lane nbr>, "indices": [<{"index": <index string>, "count": <nbr>}>]}
    """

    def __init__(self, runfolder, parser_configurations, *args, **kwargs):
        """
        Create a DemuxSummaryParser instance for the specified runfolder
        :param runfolder: path to the runfolder to parse
        :param parser_configurations: dict containing any extra configuration required by
        the parser under class name key
        """
        super().__init__(*args, **kwargs)

        self.runfolder = runfolder

        # NOTE: This parser will use the same config entry a the StatsJsonParser in order
        #       to not break backward compatibility. And it feel unnecessary to add this
        #       value to the config twice. /JD 2018-09-14
        self.parser_conf = parser_configurations.get("StatsJsonParser")
        if not self.parser_conf:
            raise ConfigurationError("The configuration must contain parser_configurations "
                                     "key with subkey StatsJsonParser. E.g: \n"
                                     "parser_configurations:\n"
                                     "\tStatsJsonParser:\n"
                                     "\t\tbcl2fastq_output_path: Data/Intensities/BaseCalls")

        self._bcl2fastq_output_path = self.parser_conf.get("bcl2fastq_output_path")
        if not self._bcl2fastq_output_path:
            raise ConfigurationError("The configuration must contain the key bcl2fastq_output_path, specifying "
                                     "where the bcl2fastq output is, relative to the runfolder root.")

        self._nbr_of_lanes = RunfolderReader.get_nbr_of_lanes(self.runfolder)
        self._validate_demux_summary_files_exist(self.runfolder, self._bcl2fastq_output_path)

    def _validate_demux_summary_files_exist(self, runfolder, bcl2fastq_output_path):
        for i in range(1, self._nbr_of_lanes):
            path = Path(runfolder, bcl2fastq_output_path, 'Stats', 'DemuxSummaryF1L{}.txt'.format(i))
            if not path.exists():
                raise DemuxSummaryNotFound("Could not identify expected demux summary file: {}. "
                                           "We expect to find {} files matching the pattern, "
                                           "'DemuxSummaryF1L<Lane number>.txt'".format(path))

    @staticmethod
    def _read_most_popular_unknown_indexes(demux_summary_file):
        with open(demux_summary_file, 'r') as f:
            reached_data = False
            for line in f:
                if reached_data:
                    split_data = line.split('\t')
                    yield {'index': split_data[0].strip(), 'count': int(split_data[1].strip())}
                if line.startswith("### Columns: Index_Sequence Hit_Count"):
                    reached_data = True

    def run(self):
        for i in range(1, self._nbr_of_lanes+1):
            path = Path(self.runfolder, self._bcl2fastq_output_path, 'Stats', 'DemuxSummaryF1L{}.txt'.format(i))
            self._send_to_subscribers(("index_counts",
                                       {"lane": i, "indices": list(self._read_most_popular_unknown_indexes(path))}))

    def __eq__(self, other):
        if isinstance(other, self.__class__) and self.runfolder == other.runfolder:
            return True
        else:
            return False

    def __hash__(self):
        return hash(self.__class__.__name__ + self.runfolder)




class Parser(object):
    """
    Parser is the base class for all parser implementations.

    A Parser will parse data and send it to its subscribers. Exactly what an item of data is depends on the particular
    implementation. The subscribers are then responsible for deciding if they are interested in that particular datum
    or not.

    Parsers should be connected to their subscribing handlers by a third class, an example of how this can
    be accomplished can be found in the QCEngine.

    Classes inheriting the Parser class must implement their `run` method. For details on what is required
    by the method implementation see the `run` method documentation.

    Furthermore in order for Parsers to be identifiable it is necessary to implement a custom version
    of `__eq__` and `__hash__`, which provides a custom definition of equivalence, this can e.g. be based on
    which runfolder the parser is setup to get its data from.
    """

    def __init__(self):
        self.subscribers = []

    def add_subscribers(self, new_subscribers):
        """
        Add the following subscriber(s) to this parser.

        :param new_subscribers: a instance of a Subscriber or a list of Subscriber instances
        :returns: None
        """
        if isinstance(new_subscribers, list):
            self.subscribers = self.subscribers + new_subscribers
        else:
            self.subscribers.append(new_subscribers)

    def _send_to_subscribers(self, value):
        """
        Calling this method will send `value` to all subscribers

        :param value: The value to send to the subscribers
        :returns: None
        """
        for subscriber in self.subscribers:
            subscriber.send(value)

    def run(self):
        """
        All Parsers must implement this method. Calling it should parse the data, what ever that means in the
        particular context, and send this data to its subscribers.

        The run method should send data to the parsers subscribers using the `_send_to_subscribers` method.

        :returns: None
        """
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError

    def __hash__(self):
        raise NotImplementedError




import json
import os
import logging

from checkQC.parsers.parser import Parser
from checkQC.exceptions import StatsJsonNotFound, ConfigurationError

log = logging.getLogger(__name__)


class StatsJsonParser(Parser):
    """
    The StatsJsonParser reads the values from the Illumina Stats.json file (which is created by bcl2fastq) and sends
    each key value pair as a tuple to the subscribers, e.g.:

        ('Flowcell', 'CB1TVANXX')
        ('RunNumber', 303)
        ('RunId', '170726_D00118_0303_BCB1TVANXX')

    The subscribers decide which of these values they are iterested in.
    """

    def __init__(self, runfolder, parser_configurations, *args, **kwargs):
        """
        Create a StatsJsonParser instance for the specified runfolder
        :param runfolder: path to the runfolder to parse
        :param parser_configurations: dict containing any extra configuration required by
        the parser under class name key
        """
        super().__init__(*args, **kwargs)

        self.parser_conf = parser_configurations.get(self.__class__.__name__)
        if not self.parser_conf:
            raise ConfigurationError("The configuration must contain parser_configurations "
                                     "key with subkey StatsJsonParser. E.g: \n"
                                     "parser_configurations:\n"
                                     "\tStatsJsonParser:\n"
                                     "\t\tbcl2fastq_output_path: Data/Intensities/BaseCalls")

        bcl2fastq_output_path = self.parser_conf.get("bcl2fastq_output_path")
        if not bcl2fastq_output_path:
            raise ConfigurationError("The configuration must contain the key bcl2fastq_output_path, specifying "
                                     "where the bcl2fastq output is, relative to the runfolder root.")

        self.file_path = os.path.join(runfolder, bcl2fastq_output_path, "Stats", "Stats.json")
        if not os.path.exists(self.file_path):
            log.error("Could not identify a Stats.json file at: {}. This file is "
                      "created by bcl2fastq, please ensure that you have run "
                      "bcl2fastq on this runfolder before running checkqc."
                      "If this file is not located under <RUNFOLDER>/Data/Intensities/BaseCalls/Stats/Stats.json "
                      "which is the default option for bcl2fastq, you can specify where the 'Stats' directory is "
                      "located by changing the 'bcl2fastq_output_path' in the 'StatsJsonParser' part of the "
                      "checkqc configuration file.".format(self.file_path))
            raise StatsJsonNotFound("Could not find a Stats.json file at: {}".format(self.file_path))

    def run(self):
        with open(self.file_path, "r") as f:
            data = json.load(f)
            for key_value in data.items():
                self._send_to_subscribers(key_value)

    def __eq__(self, other):
        if isinstance(other, self.__class__) and self.file_path == other.file_path:
            return True
        else:
            return False

    def __hash__(self):
        return hash(self.__class__.__name__ + self.file_path)




from checkQC.parsers.parser import Parser

from interop import py_interop_run_metrics, py_interop_run, py_interop_summary


class InteropParser(Parser):
    """
    This Parser will get data from the Illumina Interop binary files, and send it to its subscribers as a
    tuple with the first element being the name of the element and the second one being a the actual data.

    At this point the following data which is fetched from the Interop files and is sent in the following format:

        - ("error_rate", {"lane": <lane nbr>, "read": <read nbr>, "error_rate": <error rate>}))
        - ("percent_q30", {"lane": <lane nbr>, "read": <read nbr>, "percent_q30": <percent q30>}))

    """

    def __init__(self, runfolder, parser_configurations, *args, **kwargs):
        """
        Create a InteropParser instance for the specified runfolder

        :param runfolder: to create InteropParser instance for
        :param parser_configurations: dict containing any extra configuration required by
        the parser under class name key
        """
        super().__init__(*args, **kwargs)
        self.runfolder = runfolder

    @staticmethod
    def get_non_index_reads(summary):
        """
        Pick-out the reads which are not index reads

        :param summary: a Interop read summary object to parse the read numbers from
        :returns: all reads which are not index reads
        """
        non_index_reads = []
        for read_nbr in range(summary.size()):
            if not summary.at(read_nbr).read().is_index():
                non_index_reads.append(read_nbr)
        return non_index_reads

    def run(self):
        run_metrics = py_interop_run_metrics.run_metrics()
        run_metrics.run_info()

        valid_to_load = py_interop_run.uchar_vector(py_interop_run.MetricCount, 0)
        py_interop_run_metrics.list_summary_metrics_to_load(valid_to_load)
        run_metrics.read(self.runfolder, valid_to_load)

        summary = py_interop_summary.run_summary()
        py_interop_summary.summarize_run_metrics(run_metrics, summary)

        lanes = summary.lane_count()
        reads = self.get_non_index_reads(summary)
        for lane in range(lanes):
            # The interop library uses zero based indexing, however most people uses read 1/2
            # to denote the different reads, this enumeration is used to transform from
            # zero based indexing to this form. /JD 2017-10-27
            for new_read_nbr, original_read_nbr in enumerate(reads):
                read = summary.at(original_read_nbr).at(lane)
                error_rate = read.error_rate().mean()
                percent_q30 = read.percent_gt_q30()
                percent_phix_aligned = read.percent_aligned().mean()
                self._send_to_subscribers(("error_rate",
                                           {"lane": lane+1, "read": new_read_nbr+1, "error_rate": error_rate}))
                self._send_to_subscribers(("percent_q30",
                                           {"lane": lane+1, "read": new_read_nbr+1, "percent_q30": percent_q30}))
                self._send_to_subscribers(("percent_phix",
                                           {"lane": lane+1, "read": new_read_nbr+1, "percent_phix": percent_phix_aligned}))

    def __eq__(self, other):
        if isinstance(other, self.__class__) and self.runfolder == other.runfolder:
            return True
        else:
            return False

    def __hash__(self):
        return hash(self.__class__.__name__ + self.runfolder)




import xmltodict
import os
import logging
from checkQC.exceptions import RunParametersNotFound, RunInfoXMLNotFound

log = logging.getLogger(__name__)


class RunfolderReader(object):
    """
    This class provides methods to read files such as the runParameters.xml and RunInfo.xml files, which
    often need to be read to pick up info about what type of run we are looking at etc.
    """

    @staticmethod
    def get_nbr_of_lanes(runfolder):
        runinfo_xml = RunfolderReader.read_run_info_xml(runfolder)
        return int(runinfo_xml["RunInfo"]["Run"]["FlowcellLayout"]["@LaneCount"])

    @staticmethod
    def read_run_parameters_xml(runfolder):
        """
        Read the run parameters of an Illumina instrument are recorded in a file called
        runParameters or RunParameters depending on the exact instrument type. This method
        will read it and return it as a dict.
        :param runfolder: to look in
        :return: the [R|r]unParameters.xml as a dict
        :raises: RunParametersNotFound if no [R|r]unParameters.xml was found
        """
        try:
            with open(RunfolderReader.find_run_parameters_xml(runfolder)) as f:
                return xmltodict.parse(f.read())
        except FileNotFoundError:
            raise RunParametersNotFound("Could not find [R|r]unParameters.xml for runfolder {}".format(runfolder))

    @staticmethod
    def read_run_info_xml(runfolder):
        """
        Read the RunInfo.xml and return it as a dict
        :param runfolder: to look in
        :return: RunInfo.xml data as a dict
        """
        try:
            run_info_path = os.path.join(runfolder, "RunInfo.xml")
            if not os.path.exists(run_info_path):
                log.error("Could not find a RunInfo.xml in {}. Are you sure this is a runfolder?".format(run_info_path))
                raise FileNotFoundError("Could not find {}".format(run_info_path))
            with open(run_info_path) as f:
                return xmltodict.parse(f.read())
        except FileNotFoundError:
            raise RunInfoXMLNotFound("Could not find RunInfo.xml at {}".format(run_info_path))

    @staticmethod
    def find_run_parameters_xml(runfolder):
        """
        Finds the path to the [r|R]unParameters.xml
        :param runfolder: to look in
        :return: the path to the [r|R]unParameters.xml
        """
        first_option = os.path.join(runfolder, "RunParameters.xml")
        second_option = os.path.join(runfolder, "runParameters.xml")
        if os.path.isfile(first_option):
            return first_option
        elif os.path.isfile(second_option):
            return second_option
        else:
            log.error("Could not find [R|r]unParameters.xml in directory {}. "
                      "Are you sure this is a runfolder?".format(runfolder))
            raise FileNotFoundError("Could not find [R|r]unParameters.xml for runfolder {}".format(runfolder))




from checkQC.handlers.qc_handler import QCHandler, QCErrorFatal, QCErrorWarning
from checkQC.parsers.interop_parser import InteropParser
from checkQC.exceptions import ConfigurationError


class ErrorRateHandler(QCHandler):
    """
    This handler will check that the error rate per lane and read are below the specified threshold.
    Sometimes an error rate estimate is not available, e.g. when no PhiX has been included on
    the lane which is being analyzed. If you want to allow this to pass anyway, ensure that the
    ErrorRateHandler has 'allow_missing_error_rate' set to 'True'.
    """

    ALLOW_MISSING_ERROR_RATE = "allow_missing_error_rate"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.error_results = []

    def parser(self):
        """
        The ErrorRateHandler fetches its data from the Interop files.

        :returns: an InteropParser callable
        """
        return InteropParser

    def collect(self, signal):
        key, value = signal
        if key == "error_rate":
            self.error_results.append(value)

    def custom_configuration_validation(self):
        try:
            value = self.qc_config[self.ALLOW_MISSING_ERROR_RATE]
        except KeyError as e:
            raise ConfigurationError("Did not find key '{}' in configuration for ErrorRate handler".format(e.args[0]))

        if not isinstance(value, bool):
            raise ConfigurationError("Only True/False are allowed as values for 'allow_missing_error_rate' in the "
                                     "ErrorRate handler config. Value was: {}".format(value))

    def check_qc(self):

        for error_dict in self.error_results:
            lane_nbr = int(error_dict["lane"])
            read = error_dict["read"]
            error_rate = error_dict["error_rate"]

            if error_rate == 0 and not self.qc_config[self.ALLOW_MISSING_ERROR_RATE]:
                yield QCErrorFatal("Error rate was found to be 0 on lane: {} for read: {}, this is probably "
                                   "because there was no PhiX loaded on this lane. If do not use PhiX for your "
                                   "runs you can set 'allow_missing_error_rate' in the config to True, which will "
                                   "remove the messages in the future.".format(lane_nbr, read))
            elif self.error() != self.UNKNOWN and error_rate > self.error():
                yield QCErrorFatal("Error rate {} was to high on lane: {} for read: {}".format(error_rate,
                                                                                               lane_nbr,
                                                                                               read),
                                   ordering=lane_nbr,
                                   data={"lane": lane_nbr, "read": read,
                                         "error_rate": error_rate, "threshold": self.error()})
            elif self.warning() != self.UNKNOWN and error_rate > self.warning():
                yield QCErrorWarning("Error rate {} was to high on lane: {} for read: {}".format(error_rate,
                                                                                                 lane_nbr,
                                                                                                 read),
                                     ordering=lane_nbr,
                                     data={"lane": lane_nbr, "read": read,
                                           "error_rate": error_rate, "threshold": self.warning()})
            else:
                continue





from collections import defaultdict

from checkQC.handlers.qc_handler import QCHandler, QCErrorFatal, QCErrorWarning
from checkQC.parsers.stats_json_parser import StatsJsonParser
from checkQC.parsers.interop_parser import InteropParser


class UndeterminedPercentageHandler(QCHandler):
    """
    This handler will check that the percentage of undetermined reads on a lane is below the specified threshold.
    If there are no indexes specified for the lane, this will be skipped.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conversion_results = None
        self.phix_aligned = defaultdict(dict)

    def parser(self):
        """
        The UndeterminedPercentageHandler fetches its data from the Stats.json file

        :returns: A list of a StatsJsonParser callable and a InteropParser
        """
        return [StatsJsonParser, InteropParser]

    def collect(self, signal):
        key, value = signal
        if key == "ConversionResults":
            self.conversion_results = value
        if key == "percent_phix":
            self.phix_aligned[value["lane"]][value["read"]] = value["percent_phix"]

    def _compute_mean_percentage_phix_aligned_for_lanes(self):
        lane_and_mean_percentage_phix_aliged = {}
        for lane, reads in self.phix_aligned.items():
            mean = 0
            for read, value in reads.items():
                mean += value / len(reads)
            lane_and_mean_percentage_phix_aliged[lane] = mean
        return lane_and_mean_percentage_phix_aliged

    def check_qc(self):

        mean_phix_per_lane = self._compute_mean_percentage_phix_aligned_for_lanes()

        for lane_dict in self.conversion_results:
            # If no index was specified for a lane, there will be no
            # Undetermined key for that lane in the Stats.json file. /JD 2017-10-02
            if lane_dict.get("Undetermined"):
                lane_nbr = int(lane_dict["LaneNumber"])
                total_yield = lane_dict["Yield"]

                if total_yield == 0:
                    yield QCErrorFatal("Yield for lane: {} was 0. No undetermined percentage could be computed.",
                                       ordering=lane_nbr,
                                       data={"lane": lane_nbr, "percentage_undetermined": "N/A"})
                    continue

                undetermined_yield = lane_dict["Undetermined"]["Yield"]

                percentage_undetermined = (undetermined_yield / total_yield)*100

                def compute_threshold(value):
                    return value + mean_phix_per_lane[lane_nbr]

                def create_data_dict(value):
                    return {"lane": lane_nbr,
                            "percentage_undetermined": percentage_undetermined,
                            "threshold": value,
                            "computed_threshold": compute_threshold(value),
                            "phix_on_lane": mean_phix_per_lane[lane_nbr]}

                if self.error() != self.UNKNOWN and percentage_undetermined > compute_threshold(self.error()):
                    yield QCErrorFatal("The percentage of undetermined indexes was"
                                       " to high on lane {}, it was: {:.2f}%".format(lane_nbr,
                                                                                     percentage_undetermined),
                                       ordering=lane_nbr,
                                       data=create_data_dict(self.error()))
                elif self.warning() != self.UNKNOWN and percentage_undetermined > compute_threshold(self.warning()):
                    yield QCErrorWarning("The percentage of undetermined indexes was "
                                         "to high on lane {}, it was: {:.2f}%".format(lane_nbr,
                                                                                      percentage_undetermined),
                                         ordering=lane_nbr,
                                         data=create_data_dict(self.warning()))
                else:
                    continue




from collections import defaultdict
import re

from checkQC.handlers.qc_handler import QCHandler, QCErrorFatal, QCErrorWarning
from checkQC.parsers.demux_summary_parser import DemuxSummaryParser
from checkQC.parsers.stats_json_parser import StatsJsonParser
from checkQC.parsers.samplesheet_parser import SamplesheetParser
from checkQC.exceptions import ConfigurationError


class UnidentifiedIndexHandler(QCHandler):
    """
    The UnidentifiedIndexHandler will try to identify if an index is represented at to high a level in unidentified
    reads, and if that is the case try to pinpoint why that is.

    It will not output errors, but all information will be displayed as warnings, due to the difficulty
    of deciding what is an error or not in this context. For most cases the % of unidentified reads will be what
    is used to issue the error, and then the warnings from this handler can help in identifying the
    possible underlying cause.

    There are a number of different checks (or rules) in place, which will be checked if and index
    occurs more then the `significance_threshold`. The samplesheet will be checked to see if the index
    found matches and of the following rules:
     - Check if the dual indexes have been swapped
     - Check if the index has been reversed
     - Check if the index is the reverse complement
     - Check if the index is the complementary index
     - Check if the index is present in another lane

    It will ignore any indexes which have N's in them. These are assumed to be read errors.
    """

    WHITE_LIST_QC_KEY = 'white_listed_indexes'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lanes_and_indices = {}
        self.conversion_results = None
        self.samplesheet = None

    def validate_configuration(self):
        """
        This overrides the normal configuration which looks for warning/error.
        :return: None
        :raises: ConfigurationError if the configuration for the handler was not valid.
        """
        try:
            self.qc_config['significance_threshold']
        except KeyError:
            raise ConfigurationError("The {} handler expects 'significance_threshold' to be set. "
                                     "Perhaps it is missing from the configuration "
                                     "file?".format(self.__class__.__name__))

    def parser(self):
        """
        The UnidentifiedIndexHandler fetches its information from the
        :return:
        """
        return [DemuxSummaryParser, StatsJsonParser, SamplesheetParser]

    def collect(self, signal):
        key, value = signal
        if key == "ConversionResults":
            self.conversion_results = value
        if key == "samplesheet":
            self.samplesheet = value
        if key == "index_counts":
            self.lanes_and_indices[value["lane"]] = value["indices"]

    def check_qc(self):

        samplesheet_searcher = _SamplesheetSearcher(samplesheet=self.samplesheet)
        number_of_reads_per_lane = self.number_of_reads_per_lane()

        for lane, indices in self.lanes_and_indices.items():
            number_of_reads_on_lane = number_of_reads_per_lane[lane]
            for index in indices:
                tag = index['index']
                count = index['count']
                if self.should_be_evaluated(tag, count, number_of_reads_on_lane):
                    percent_on_lane = (float(count) / number_of_reads_on_lane) * 100
                    yield from self.evaluate_index_rules(tag, lane, samplesheet_searcher, percent_on_lane)
                else:
                    continue

    def should_be_evaluated(self, tag, count, number_of_reads_on_lane):
        # Unknown index means that the sample was run without an index.
        # Ns indicate errors in read
        return not tag == 'unknown' and \
               self.is_significantly_represented(count, number_of_reads_on_lane)

    def _tag_in_white_list(self, tag):
        # This preserves backward compatiblity for users that do not specify
        # the 'white_listed_indexes' in their configs. /JD 2019-06-12
        if not self.qc_config.get(self.WHITE_LIST_QC_KEY):
            return 'N' in tag
        else:
            for regex in self.qc_config[self.WHITE_LIST_QC_KEY]:
                pattern = re.compile(regex)
                if pattern.match(tag):
                    return True
            return False

    def is_significantly_represented(self, index_count, nbr_of_reads_on_lane):
        return (float(index_count) / nbr_of_reads_on_lane) > \
               (float(self.qc_config['significance_threshold']) / 100)

    def evaluate_index_rules(self, tag, lane, samplesheet_searcher, percent_on_lane):
        """
        Evaluates a list of 'rules' and yields all warnings found by these rules.
        :param tag:
        :param lane:
        :param samplesheet_searcher:
        :param percent_on_lane:
        :return: generator of QCErrorFatal
        """
        rules = [self.always_warn_rule, self.check_reversed_index,
                 self.check_reverse_complement_index, self.check_if_index_in_other_lane,
                 self.check_complement_index]
        for rule in rules:
            yield from rule(tag=tag,
                            lane=lane,
                            samplesheet_searcher=samplesheet_searcher,
                            percent_on_lane=percent_on_lane)

        dual_index_rules = [self.check_swapped_dual_index, self.check_reverse_complement_in_dual_index,
                            self.check_reversed_in_dual_index, self.check_complement_in_dual_index]

        if '+' in tag:
            for rule in dual_index_rules:
                yield from rule(tag=tag,
                                lane=lane,
                                samplesheet_searcher=samplesheet_searcher,
                                percent_on_lane=percent_on_lane)

    @staticmethod
    def yield_qc_warning_with_message(msg, lane, tag):
        yield QCErrorFatal(msg=msg, ordering="{}:{}".format(lane, tag), data={'lane': lane, 'msg': msg})

    def always_warn_rule(self, tag, lane, percent_on_lane, **kwargs):
        """
        We always want to warn about an index that is significantly represented. This rule
        will make sure that we do so, and all other rules will contribute extra information
        if there is any.
        :param tag:
        :param lane:
        :param percent_on_lane:
        :return:
        """
        if self._tag_in_white_list(tag):
            msg = ("Index: {} on lane: {} was significantly overrepresented ({:.1f}%) at a " +
                   "significance threshold of {}%. " +
                   "This index is white-listed.").format(tag,
                                                         lane,
                                                         percent_on_lane,
                                                         self.qc_config["significance_threshold"])
            yield QCErrorWarning(msg=msg, ordering="{}:{}".format(lane, tag), data={'lane': lane, 'msg': msg})
        else:
            yield from UnidentifiedIndexHandler.\
                yield_qc_warning_with_message("Index: {} on lane: {} was significantly "
                                              "overrepresented ({:.1f}%) at significance "
                                              "threshold of: {}%.".format(tag,
                                                                          lane,
                                                                          percent_on_lane,
                                                                          self.qc_config["significance_threshold"]),
                                              lane,
                                              tag)

    @staticmethod
    def check_swapped_dual_index(tag, lane, samplesheet_searcher, **kwargs):
        if '+' in tag:
            split_tag = tag.split('+')
            swapped_tag = '{}+{}'.format(split_tag[1], split_tag[0])
            hits = samplesheet_searcher.exact_index_in_samplesheet(swapped_tag)
            for hit in hits:
                msg = '\tIt appears that maybe the dual index tag: {} was swapped. There was a hit for' \
                      ' the swapped index: {} at: {}'.format(tag, swapped_tag, hit)
                yield from UnidentifiedIndexHandler.yield_qc_warning_with_message(msg, lane, tag)

    @staticmethod
    def check_reversed_in_dual_index(tag, lane, samplesheet_searcher, **kwargs):
        if '+' in tag:
            split_tags = tag.split('+')
            for single_tag in split_tags:
                hits = samplesheet_searcher.one_index_match_from_dual_index_in_samplesheet(single_tag[::-1])
                for hit in hits:
                    msg = '\tWe found a possible match for the reverse of tag: {}, on: {}. '.format(single_tag, hit)
                    yield from UnidentifiedIndexHandler.yield_qc_warning_with_message(msg, lane, tag)

    @staticmethod
    def check_reverse_complement_in_dual_index(tag, lane, samplesheet_searcher, **kwargs):
        if '+' in tag:
            split_tags = tag.split('+')
            for single_tag in split_tags:
                hits = samplesheet_searcher.one_index_match_from_dual_index_in_samplesheet(
                    UnidentifiedIndexHandler.get_complementary_sequence(single_tag)[::-1])
                for hit in hits:
                    msg = '\tWe found a possible match for the reverse complement of tag: {}, on: {}. '.format(single_tag, hit)
                    yield from UnidentifiedIndexHandler.yield_qc_warning_with_message(msg, lane, tag)

    @staticmethod
    def check_complement_in_dual_index(tag, lane, samplesheet_searcher, **kwargs):
        if '+' in tag:
            split_tags = tag.split('+')
            for single_tag in split_tags:
                hits = samplesheet_searcher.exact_index_in_samplesheet(
                    UnidentifiedIndexHandler.get_complementary_sequence(single_tag))
                for hit in hits:
                    msg = '\tWe found a possible match for the complement of tag: {}, on: {}. '.format(single_tag, hit)
                    yield from UnidentifiedIndexHandler.yield_qc_warning_with_message(msg, lane, tag)

    @staticmethod
    def check_reversed_index(tag, lane, samplesheet_searcher, **kwargs):
        hits = samplesheet_searcher.exact_index_in_samplesheet(tag[::-1])
        for hit in hits:
            msg = '\tWe found a possible match for the reverse of tag: {}, on: {}'.format(tag, hit)
            yield from UnidentifiedIndexHandler.yield_qc_warning_with_message(msg, lane, tag)

    @staticmethod
    def check_reverse_complement_index(tag, lane, samplesheet_searcher, **kwargs):
        hits = samplesheet_searcher.exact_index_in_samplesheet(
            UnidentifiedIndexHandler.get_complementary_sequence(tag)[::-1])
        for hit in hits:
            msg = '\tWe found a possible match for the reverse complement of tag: {}, on: {}'.format(tag, hit)
            yield from UnidentifiedIndexHandler.yield_qc_warning_with_message(msg, lane, tag)

    @staticmethod
    def check_complement_index(tag, lane, samplesheet_searcher, **kwargs):
        hits = samplesheet_searcher.exact_index_in_samplesheet(
            UnidentifiedIndexHandler.get_complementary_sequence(tag))
        for hit in hits:
            msg = '\tWe found a possible match for the complementary of tag: {}, on: {}'.format(tag, hit)
            yield from UnidentifiedIndexHandler.yield_qc_warning_with_message(msg, lane, tag)

    @staticmethod
    def check_if_index_in_other_lane(tag, lane, samplesheet_searcher, **kwargs):
        hits = samplesheet_searcher.exact_index_in_samplesheet(tag)
        for hit in hits:
            msg = '\tWe found a possible match for the tag: {}, on another lane: {}'.format(tag, hit)
            yield from UnidentifiedIndexHandler.yield_qc_warning_with_message(msg, lane, tag)

    def number_of_reads_per_lane(self):
        """
        Transform conversion results into dict of lane -> total clusters pass filer
        :return: dict {<lane>: <total clusters pass filer>}
        """
        nbr_of_reads_per_lane = {}
        for lane_dict in self.conversion_results:
            nbr_of_reads_per_lane[int(lane_dict["LaneNumber"])] = int(lane_dict["TotalClustersPF"])
        return nbr_of_reads_per_lane

    @staticmethod
    def get_complementary_sequence(seq):
        conv_table = {'A': 'T', 'T': 'A', 'G': 'C', 'C': 'G', 'N': 'N', '+': '+'}
        comp_seq = ''
        for c in seq:
            comp_seq += conv_table[c]
        return comp_seq


class _SamplesheetSearcher(object):

    class SearchHit(object):
        def __init__(self, search_index, found_sample, found_lane):
            self.search_index = search_index
            self.found_sample = found_sample
            self.found_lane = found_lane

        def __str__(self):
            return "Lane: {}, for sample: {}. The tag we found in the samplesheet was: {}".format(self.found_lane,
                                                                               self.found_sample,
                                                                               self.search_index)

        def __eq__(self, other):
            if isinstance(other, self.__class__) and self.search_index == other.search_index \
                    and self.found_sample == other.found_sample and self.found_lane == other.found_lane:
                return True
            else:
                return False

        def __hash__(self):
            return hash(self.__class__.__name__ + self.search_index + self.found_sample + self.found_lane)

    def __init__(self, samplesheet):
        self.samplesheet_dict = self.transform_samplesheet_to_dict(samplesheet)

    @staticmethod
    def transform_samplesheet_to_dict(samplesheet):
        """
        Transform samplesheet to dict on form {index -> {lane -> sample name}}
        :param samplesheet:
        :return: dict of index, lane and sample name
        """

        samplesheet_dict = defaultdict(lambda: defaultdict(dict))
        for s in samplesheet:
            index = s['index']
            if s.get('index2'):
                index += '+{}'.format(s['index2'])
            # MiSeqs for example do not have lanes in their samplesheets, thus we will default to using lane 1
            # if nothing has been specified in the samplesheet. /JD 20181101
            lane = s.get('Lane', '1')
            samplesheet_dict[index][lane] = s['Sample_Name']

        return samplesheet_dict

    def exact_index_in_samplesheet(self, index):
        """
        Search the sample sheet dict for an index to find which lane and which sample
        it is associated with.
        :param index:
        :return: generator of SearchHits
        """
        index_hits = self.samplesheet_dict.get(index)
        if index_hits:
            for lane, sample in index_hits.items():
                yield _SamplesheetSearcher.SearchHit(index, sample, lane)

    def one_index_match_from_dual_index_in_samplesheet(self, index):
        """
        Search for in-exact matches in samplesheet, i.e. hits where there is a hit, for one, but not
        both indexes in a dual index tag.
        :param index:
        :return:
        """

        for samplesheet_index in self.samplesheet_dict.keys():
            if index in set(samplesheet_index.split('+')):
                for lane, sample in self.samplesheet_dict[samplesheet_index].items():
                    yield _SamplesheetSearcher.SearchHit(index, sample, lane)




import logging
import logging.config
import os

import click
import json
from json.decoder import JSONDecodeError

import tornado.ioloop
import tornado.web
from tornado.web import url, HTTPError

from checkQC.app import App
from checkQC.config import ConfigFactory
from checkQC.exceptions import *

log = logging.getLogger(__name__)

from checkQC import __version__ as checkqc_version


class CheckQCHandler(tornado.web.RequestHandler):

    def initialize(self, **kwargs):
        self.monitor_path = kwargs["monitoring_path"]
        self.qc_config_file = kwargs["qc_config_file"]
        self.downgrade_errors_for = ()
        self.use_closest_read_length = False

    @staticmethod
    def _run_check_qc(monitor_path, qc_config_file, runfolder, downgrade_errors_for,
                      use_closest_read_length):
        path_to_runfolder = os.path.join(monitor_path, runfolder)
        checkqc_app = App(config_file=qc_config_file, runfolder=path_to_runfolder,
                          downgrade_errors_for=downgrade_errors_for,
                          use_closest_read_length=use_closest_read_length)
        reports = checkqc_app.configure_and_run()
        reports["version"] = checkqc_version
        return reports

    def _write_error(self, status_code, reason):
        self.set_header("Content-Type", "application/json")
        self.set_status(status_code=status_code)
        self.finish({"reason": reason})

    def get(self, runfolder):
        if "downgrade" in self.request.query_arguments:
            self.downgrade_errors_for = self.get_query_argument("downgrade")
        if "useClosestReadLength" in self.request.query_arguments:
            self.use_closest_read_length = True
        try:
            reports = self._run_check_qc(self.monitor_path, self.qc_config_file,
                                         runfolder, self.downgrade_errors_for,
                                         self.use_closest_read_length)
            self.set_header("Content-Type", "application/json")
            self.write(reports)
        except RunfolderNotFoundError:
            self._write_error(status_code=404, reason="Could not find requested runfolder.")
        except ConfigurationError:
            self._write_error(status_code=500, reason="There is a problem with the qc config. Are you sure the "
                                                      "type of instrument/run configuration on the run you want to "
                                                      "analyze is available in the qc config?")


class WebApp(object):

    def __init__(self):
        pass

    @staticmethod
    def _routes(**kwargs):
        return [url(r"/qc/([^/]+)", CheckQCHandler, name="checkqc", kwargs=kwargs)]

    @staticmethod
    def _make_app(debug=False, **kwargs):
        return tornado.web.Application(WebApp._routes(**kwargs), debug=debug)

    def start_web_app(self, monitoring_path, port, config_file, log_config, debug):
        logging_config_path = ConfigFactory.get_logging_config_dict(log_config)
        logging.config.dictConfig(logging_config_path)

        log.info("Starting checkqc-ws at port: {}".format(port))

        if not os.path.isdir(monitoring_path):
            log.error("{} is not a directory".format(monitoring_path))
            raise AssertionError("{} is not a directory".format(monitoring_path))

        web_app = self._make_app(monitoring_path=monitoring_path, qc_config_file=config_file, debug=debug)
        web_app.listen(port=port)
        tornado.ioloop.IOLoop.instance().start()


@click.command("checkqc-ws")
@click.argument('monitor_path', type=click.Path())
@click.option("--port", help="Port which checkqc-ws will listen to (default: 9999).", type=click.INT, default=9999)
@click.option("--config", help="Path to the checkQC configuration file (optional)", type=click.Path())
@click.option("--log_config", help="Path to the checkQC logging configuration file (optional)", type=click.Path())
@click.option('--debug', is_flag=True, default=False, help="Enable debug mode.")
def start(monitor_path, port=9999, config=None, log_config=None, debug=False):
    webapp = WebApp()
    webapp.start_web_app(monitor_path, port, config, log_config, debug)




import logging

from checkQC.exceptions import *
from checkQC.runfolder_reader import RunfolderReader

log = logging.getLogger(__name__)


class IlluminaInstrument(object):
    """
    Base class representing an Illumina instrument. The `name` and `reagent_version` needs to be implemented
    by the specific subclasses.
    """

    @staticmethod
    def get_subclasses():
        """
        Get all subclasses which extends this class (which should be all supported Illumina instruments)

        :returns: a list of IlluminaInstrument subclasses
        """
        return IlluminaInstrument.__subclasses__()

    @staticmethod
    def create_instrument_instance(instrument_name):
        """
        Get the instrument instance corresponding to the given instrument name

        :param instrument_name: name of instrument to get the implementing class for
        :returns: a instance of the corresponding IlluminaInstrument
        """
        subclasses = IlluminaInstrument.get_subclasses()
        for subclass in subclasses:
            if instrument_name == subclass.name():
                return subclass()
        raise InstrumentTypeUnknown

    @staticmethod
    def name():
        """
        Name of the instrument, e.g. 'nova_seq'

        :returns: name of instrument as string
        """
        raise NotImplementedError

    @staticmethod
    def reagent_version(runtype_recognizer):
        """
        Reagent version, e.g. `v1`
        Can used the provided runtype_recognizer to determined the exact reagent version

        :param runtype_recognizer: A instance of RuntypeRecognizer
        :returns: reagent version as a string
        """
        raise NotImplementedError

class ISeq(IlluminaInstrument):

    @staticmethod
    def name():
        return "iseq"

    @staticmethod
    def reagent_version(runtype_recognizer):
        return "v1"

class NovaSeq(IlluminaInstrument):

    @staticmethod
    def name():
        return "novaseq"

    @staticmethod
    def reagent_version(runtype_recognizer):
        try:
            reagent_version = runtype_recognizer.run_parameters["RunParameters"]["RfidsInfo"]["FlowCellMode"]
            return reagent_version
        except KeyError:
            raise ReagentVersionUnknown("Could not identify flowcell mode for Novaseq")


class HiSeqX(IlluminaInstrument):

    @staticmethod
    def name():
        return "hiseqx"

    @staticmethod
    def reagent_version(runtype_recognizer):
        return "v2"


class MiSeq(IlluminaInstrument):

    @staticmethod
    def name():
        return "miseq"

    @staticmethod
    def reagent_version(runtype_recognizer):
        """
        Find the reagent version used for this run, as MiSeqs can have multiple
        different reagent kit versions.

        :returns: reagent version of format v[number] e.g. v3
        """
        try:
            reagent_version = runtype_recognizer.run_parameters["RunParameters"]["ReagentKitVersion"]
            return reagent_version.replace("Version", "v")
        except KeyError:
            raise ReagentVersionUnknown("No reagent version specified for this instrument type")


class HiSeq2500(IlluminaInstrument):

    @staticmethod
    def name():
        return "hiseq2500"

    @staticmethod
    def reagent_version(runtype_recognizer):
        """
        Find run mode (rapid or not) and reagent version used for this run

        :return run mode (as specified in RunInfo.xml) and reagent version
                joint as one string e.g. rapidhighoutput_v4 or rapidrun_v2
        """
        try:
            run_mode = runtype_recognizer.run_parameters["RunParameters"]["Setup"]["RunMode"].lower()
        except KeyError:
            raise RunModeUnknown("No run mode specified for this instrument type")

        try:
            reagent_version = runtype_recognizer.run_parameters["RunParameters"]["Setup"]["Sbs"]
            #Select last element from string "HiSeq SBS Kit v4"
            format_reagent_version= reagent_version.split(" ")[-1].strip().lower()
        except KeyError:
            raise ReagentVersionUnknown("No reagent version specified for this instrument type and run mode")

        return "{}_{}".format(run_mode, format_reagent_version)


class RunTypeRecognizer(object):
    """
    RunTypeRecognizer will read files in the runfolder to determine information about the run,
    such as the instrument type, the read length, etc.

    The runfolder needs to have a 'RunInfo.xml' and a '[R|r]unParameters.xml' file.
    """

    def __init__(self, runfolder, runfolder_reader=RunfolderReader()):
        """
        Create a RunTypeRecognizer instance

        :param runfolder: to gather data about
        :param runfolder_reader: reader class for for runfolders, defaults to RunfolderReader. Here to make testing
        easier.
        """
        self._runfolder = runfolder
        self.run_info = runfolder_reader.read_run_info_xml(runfolder)
        self.run_parameters = runfolder_reader.read_run_parameters_xml(runfolder)


    def instrument_type(self):
        """
        This will look in the RunInfo.xml and determine the run type, based on the
        mappings from instrument names to instrument types

        :raises: InstrumentTypeUnknown
        :returns: the instrument type of the runfolder
        """
        instrument_name = self.run_info["RunInfo"]["Run"]["Instrument"]
        machine_type_mappings = {"M": "miseq",
                                 "D": "hiseq2500",
                                 "ST": "hiseqx",
                                 "A": "novaseq",
                                 "FS": "iseq"}

        for key, value in machine_type_mappings.items():
            if instrument_name.startswith(key):
                return IlluminaInstrument.create_instrument_instance(value)

        raise InstrumentTypeUnknown("Did not recognize instrument type of: {}".format(instrument_name))

    def instrument_and_reagent_version(self):
        """
        Get the instrument and reagent version associated with this runfolder.

        :returns: the joined instrument and reagent version, e.g. 'hiseq2500_rapidrun_v2'
        """
        instrument_type = self.instrument_type()
        return "_".join([instrument_type.name(), instrument_type.reagent_version(self)])

    def read_length(self):
        """
        Gather information on the read length of the run.

        :returns: The read length. If multiple reads delimited by "-", e.g. 150-150.
        """
        reads = self.run_info["RunInfo"]["Run"]["Reads"]["Read"]

        read_lengths = []
        for read in reads:
            if not read['@IsIndexedRead'] == 'Y':
                read_lengths.append(int(read['@NumCycles']))

        if len(read_lengths) < 1:
            raise RunModeUnknown("Found no NumCycles in RunInfo.xml, could not determine read length")

        return "-".join(map(str, read_lengths))







from checkQC.handlers.qc_handler import QCHandler, QCErrorFatal, QCErrorWarning
from checkQC.parsers.interop_parser import InteropParser


class Q30Handler(QCHandler):
    """
    This handler checks that the percent of bases on a lane and reads with Q30 or high was
    above the specified threshold.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.error_results = []

    def parser(self):
        """
        The Q30Handler fetches its data from the Interop files

        :returns: A InteropParser callable
        """
        return InteropParser

    def collect(self, signal):
        key, value = signal
        if key == "percent_q30":
            self.error_results.append(value)

    def check_qc(self):

        for error_dict in self.error_results:
            lane_nbr = int(error_dict["lane"])
            read = error_dict["read"]
            percent_q30 = error_dict["percent_q30"]

            if self.error() != self.UNKNOWN and percent_q30 < self.error():
                yield QCErrorFatal("%Q30 {:.2f} was too low on lane: {} for read: {}".format(percent_q30,
                                                                                             lane_nbr,
                                                                                             read),
                                   ordering=lane_nbr,
                                   data={"lane": lane_nbr, "read": read,
                                         "percent_q30": percent_q30, "threshold": self.error()})
            elif self.warning() != self.UNKNOWN and percent_q30 < self.warning():
                yield QCErrorWarning("%Q30 {:.2f} was too low on lane: {} for read: {}".format(percent_q30,
                                                                                               lane_nbr,
                                                                                               read),
                                     ordering=lane_nbr,
                                     data={"lane": lane_nbr, "read": read,
                                           "percent_q30": percent_q30, "threshold": self.warning()})
            else:
                continue




from checkQC.handlers.qc_handler import QCHandler, QCErrorFatal, QCErrorWarning
from checkQC.parsers.stats_json_parser import StatsJsonParser
from math import pow
from collections import defaultdict


class ReadsPerSampleHandler(QCHandler):
    """
    This handler will check that the number of reads assigned to a sample is high enough. The value specified in the
    configuration is interpreted as the number of reads demanded for a single sample, i.e. the number of reads per
    sample on a lane which has multiple samples is the threshold divided by the total number of samples on the lane.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conversion_results = None

    def parser(self):
        """
        The ReadsPerSampleHandler fetches its information from the Stats.json file

        :returns: A StatsJsonParser callable
        """
        return StatsJsonParser

    def collect(self, signal):
        key, value = signal
        if key == "ConversionResults":
            self.conversion_results = value

    def check_qc(self):

        for lane_dict in self.conversion_results:
            lane_nbr = int(lane_dict["LaneNumber"])
            lane_demux = lane_dict["DemuxResults"]
            total_reads = defaultdict(float)

            for sample_id_info in lane_demux:
                sample_name = sample_id_info["SampleName"]
                total_reads[sample_name] += sample_id_info["NumberReads"] / pow(10, 6)

            nbr_of_samples = len(total_reads.keys())
            for sample, sample_total_reads in total_reads.items():

                if self.error() != self.UNKNOWN:
                    error_threshold = float(self.error()) / float(nbr_of_samples)
                if self.warning() != self.UNKNOWN:
                    warning_threshold = float(self.warning()) / float(nbr_of_samples)

                if self.error() != self.UNKNOWN and sample_total_reads < error_threshold:
                    yield QCErrorFatal("Number of reads for sample {} was too low on lane {}, "
                                       "it was: {:.3f} M".format(sample, lane_nbr, sample_total_reads),
                                       ordering=lane_nbr,
                                       data={"lane": lane_nbr, "number_of_samples": nbr_of_samples,
                                             "sample_name": sample, "sample_reads": sample_total_reads,
                                             "threshold": error_threshold})
                elif self.warning() != self.UNKNOWN and \
                                sample_total_reads < warning_threshold:
                    yield QCErrorWarning("Number of reads for sample {} was too low on lane {}, "
                                         "it was: {:.3f} M".format(sample, lane_nbr, sample_total_reads),
                                         ordering=lane_nbr,
                                         data={"lane": lane_nbr, "number_of_samples": nbr_of_samples,
                                               "sample_name": sample, "sample_reads": sample_total_reads,
                                               "threshold": warning_threshold})
                else:
                    continue



from math import pow

from checkQC.handlers.qc_handler import QCHandler, QCErrorFatal, QCErrorWarning
from checkQC.parsers.stats_json_parser import StatsJsonParser


class ClusterPFHandler(QCHandler):
    """
    This handler will check that the number of clusters passing filter on a lane passes the set criteria.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conversion_results = None

    def parser(self):
        """
        The ClusterPFHandler will gather its data from the Stats.json file

        :returns: a StatsJsonParser callable
        """
        return StatsJsonParser

    def collect(self, signal):
        key, value = signal
        if key == "ConversionResults":
            self.conversion_results = value

    def check_qc(self):
        for lane_dict in self.conversion_results:
            lane_nbr = int(lane_dict["LaneNumber"])
            lane_pf = lane_dict["TotalClustersPF"]

            if self.error() != self.UNKNOWN and lane_pf < float(self.error())*pow(10, 6):
                yield QCErrorFatal("Clusters PF was to low on lane {}, "
                                   "it was: {:.2f} M".format(lane_nbr, lane_pf/pow(10, 6)),
                                   ordering=lane_nbr,
                                   data={'lane': lane_nbr, 'lane_pf': lane_pf, 'threshold': self.error()})
            elif self.warning() != self.UNKNOWN and lane_pf < float(self.warning())*pow(10, 6):
                yield QCErrorWarning("Cluster PF was to low on lane {}, "
                                     "it was: {:.2f} M".format(lane_nbr, lane_pf/pow(10, 6)),
                                     ordering=lane_nbr,
                                     data={'lane': lane_nbr, 'lane_pf': lane_pf, 'threshold': self.warning()})
            else:
                continue



from setuptools import setup, find_packages
from checkQC import __version__

setup(
    name='checkQC',
    version=__version__,
    description="A simple program to parse Illumina NGS data and check it for quality criteria.",
    long_description="A simple program to parse Illumina NGS data and check it for quality criteria.",
    keywords=['bioinformatics', 'ngs', 'quality control'],
    author='Johan Dahlberg, SNP&SEQ Technology Platform, Uppsala University',
    author_email='johan.dahlberg@medsci.uu.se',
    url="https://www.github.com/Molmed/checkQC",
    download_url='https://github.com/Molmed/checkQC/archive/{}.tar.gz'.format(__version__),
    install_requires=[
        "click",
        "PyYAML>=3.12",
        "interop>=1.1.10",
        "xmltodict",
        "tornado",
        "sample_sheet"],
    packages=find_packages(exclude=["tests*"]),
    test_suite="tests",
    package_data={'checkQC': ['default_config/config.yaml', 'default_config/logger.yaml']},
    include_package_data=True,
    license='GPLv3',
    entry_points={
        'console_scripts': ['checkqc = checkQC.app:start',
                            'checkqc-ws = checkQC.web_app:start']
    },
)




import importlib
import pkgutil

import checkQC.handlers
from checkQC.handlers.qc_handler import QCHandler
from checkQC.exceptions import QCHandlerNotFound


class QCHandlerFactory(object):
    """
    This class provides way of finding and instantiating a concrete QCHandler implementation.
    This allows QCHandlers to be instantiated dynamically at runtime e.g. based on what is
    specified in a config file.
    """

    @staticmethod
    def create_subclass_instance(class_name, class_config):
        """
        This method will look for a class with the given `class_name` in the `checkQC.handlers` module.
        If it can find a class with a matching name it will return a instance of that class.

        :param class_name: the name of the class to instantiate
        :param class_config: dictionary with configuration for the class
        :returns: A instance of the class represented by class_name
        """
        package = checkQC.handlers
        prefix = package.__name__ + "."

        modules = list(pkgutil.iter_modules(package.__path__, prefix))
        for importer, modname, ispkg in modules:
            importlib.import_module(modname)
        qc_handler_subclasses = list(QCHandler.__subclasses__())
        try:
            i = list(map(lambda clazz: clazz.__name__, qc_handler_subclasses)).index(class_name)
            return qc_handler_subclasses[i](qc_config=class_config)
        except ValueError:
            raise QCHandlerNotFound("Could not identify a QCHandler with name: {}".format(class_name))





from pathlib import Path

from sample_sheet import SampleSheet

from checkQC.parsers.parser import Parser
from checkQC.exceptions import ConfigurationError, SamplesheetNotFound


class SamplesheetParser(Parser):
    """
    TODO
    """

    def __init__(self, runfolder, parser_configurations, *args, **kwargs):
        """
        Create a SamplesheetParser instance for the specified runfolder
        :param runfolder: path to the runfolder to parse
        :param parser_configurations: dict containing any extra configuration required by
        the parser under class name key
        """
        super().__init__(*args, **kwargs)

        self.runfolder = runfolder

        self.parser_conf = parser_configurations.get(self.__class__.__name__)
        if not self.parser_conf:
            raise ConfigurationError("The configuration must contain parser_configurations "
                                     "key with subkey SamplesheetParser. E.g: \n"
                                     "parser_configurations:\n"
                                     "\tSamplesheetParser:\n"
                                     "\t\tsamplesheet_name: Samplesheet.csv")

        self._samplesheet_name = self.parser_conf.get("samplesheet_name")
        if not self._samplesheet_name:
            raise ConfigurationError("The configuration must contain the key samplesheet_name, specifying "
                                     "what the name of the samplesheet is.")

        self._samplesheet = Path(self.runfolder, self._samplesheet_name)
        if not self._samplesheet.exists():
            raise SamplesheetNotFound("Could not identify samplesheet at: {}".format(self._samplesheet))

    def run(self):
        samplesheet_read = SampleSheet(self._samplesheet)
        self._send_to_subscribers(("samplesheet", samplesheet_read))

    def __eq__(self, other):
        if isinstance(other, self.__class__) and self.runfolder == other.runfolder:
            return True
        else:
            return False

    def __hash__(self):
        return hash(self.__class__.__name__ + self.runfolder)




from pkg_resources import Requirement, resource_filename
import logging
from checkQC.exceptions import ConfigEntryMissing

import yaml


log = logging.getLogger(__name__)


class ConfigFactory(object):
    """
    The ConfigFactory provides methods for creating a Config instance.
    """

    @staticmethod
    def from_config_path(config_path):
        """
        Creates a Config instance from the provided config_path

        :param config_path: path to the configuration, or None. If no config_path is provided the default config file will be used \
        :returns: Config instance based on the specified file path
        """
        config_file_content = ConfigFactory._get_config_file(config_path)
        return Config(config_file_content)

    @staticmethod
    def _get_config_file(config_path):
        """
        Load the content of the config file. If no config_path is specified, get the default of config file.

        :param config_path: path to the config file or None
        :returns: the content of the config file
        """
        try:
            if not config_path:
                config_path = resource_filename(Requirement.parse('checkQC'), 'checkQC/default_config/config.yaml')
                log.info("No config file specified, using default config from {}.".format(config_path))

            with open(config_path) as stream:
                return yaml.load(stream)
        except FileNotFoundError as e:
            log.error("Could not find config file: {}".format(e))
            raise e

    @staticmethod
    def get_logging_config_dict(config_path):
        """
        Loads the specified logger config. This is useful when CheckQC is used more like a library, so that the
        default logging configuration can be overridden.

        :param config_path: Path to the logging config.
        :returns: The content of the logging config file.
        """
        try:
            if not config_path:
                config_path = resource_filename(Requirement.parse('checkQC'), 'checkQC/default_config/logger.yaml')
                log.info("No logging config file specified, using default config from {}.".format(config_path))
            with open(config_path) as stream:
                return yaml.load(stream)
        except FileNotFoundError as e:
            log.error("Could not find config file: {}".format(e))
            raise e


class Config(object):
    """
    A Config object wraps the configuration for all handlers, so that the correct config can be passed to a handler
    depending on e.g. which read length and run type has been used for a sequencing run.
    """

    def __init__(self, config):
        """
        Create a Config instance.

        :param config: content of the config file
        """
        self._config = config

    def _get_matching_handler(self, instrument_and_reagent_type, read_length, use_closest_read_length=False):
        """
        Get the handler matching the provided parameters.

        :param instrument_and_reagent_type: the instrument and run type, e.g. 'hiseq2500_rapidhighoutput_v4'
        :param read_length: either as a range, e.g. '50-70' or a single value, e.g. '50'
        :returns: A dict corresponding to the handler config
        :raises: ConfigEntryMissing if instrument, reagent type and read length detected is missing from config
        """

        try:
            config_read_lengths = list(map(str, self._config[instrument_and_reagent_type].keys()))
            for config_read_length in config_read_lengths:
                if "-" in config_read_length:
                    split_read_length = config_read_length.split("-")
                    low_break = int(split_read_length[0])
                    high_break = int(split_read_length[1])
                    if low_break <= int(read_length) <= high_break:
                        return self._config[instrument_and_reagent_type][config_read_length]["handlers"]
                else:
                    if int(read_length) == int(config_read_length):
                        return self._config[instrument_and_reagent_type][int(config_read_length)]["handlers"]
            if use_closest_read_length:
                closest_read_length = self._find_closest_read_length(config_read_lengths, read_length)
                log.info(f"Read length {read_length} not find in config. Using closest read length: {closest_read_length}")
                return self._config[instrument_and_reagent_type][closest_read_length]["handlers"]
            raise ConfigEntryMissing("Could not find a config entry matching read length '{}' on "
                                     "instrument '{}'. Please check the provided "
                                     "config.".format(read_length, instrument_and_reagent_type))
        except KeyError:
            raise ConfigEntryMissing("Could not find a config entry for instrument '{}' "
                                     "with read length '{}'. Please check the provided config "
                                     "file ".format(instrument_and_reagent_type,
                                                    read_length))

    def _find_closest_read_length(self, config_read_lengths, read_length):
        """
        Find the closest read length in the config

        :param config_read_lengths: dict with config read lengths for a specific intrument and reagent type
        :param read_length:  either as a range, e.g. '50-70' or a single value, e.g. '50'
        :returns: the closest read length, as a string (if interval) or int (if single value)
        """
        distance = {}
        for config_read_length in sorted(config_read_lengths, reverse=True):
            if "-" in config_read_length:
                split_read_length = config_read_length.split("-")
                distance[config_read_length] = min(abs(int(read_length) - int(split_read_length[0])),
                                                   abs(int(read_length) - int(split_read_length[1])))
            else:
                distance[config_read_length] = abs(int(read_length) - int(config_read_length))
        closest_read_length = min(distance, key=distance.get)
        if "-" not in closest_read_length:
            return int(closest_read_length)
        else:
            return closest_read_length

    def _add_default_config(self, current_handler_config):
        """
        Add the default handlers specified in the config.

        :param current_handler_config: a list of handlers. This will be mutated.
        :returns: The provided list with the default configs added to it.
        """

        current_handler_names = set(map(lambda x: x["name"], current_handler_config))
        default_handlers = self._config["default_handlers"]
        for default_handler in default_handlers:
            if not default_handler["name"] in current_handler_names:
                current_handler_config.append(default_handler)
        return current_handler_config

    def _downgrade_errors(self, current_handler_config, downgrade_errors_for):
        """
        Downgrade errors to warnings for specific handlers

        :param current_handler_config: a list of handlers.
        :param downgrade_errors_for: a tuple of handlers for which errors should be downgraded
        :returns: The provided list where some handlers might have been modified
        """
        downgraded_handler_config = []
        for handler in current_handler_config:
            if handler["name"] in downgrade_errors_for and handler["error"] != "unknown":
                log.info("Downgrading errors for {}".format(handler["name"]))
                handler["warning"] = handler["error"]
                handler["error"] = "unknown"
                downgraded_handler_config.append(handler)
            else:
                downgraded_handler_config.append(handler)
        return downgraded_handler_config

    def get_handler_configs(self, instrument_and_reagent_type, read_length,
                            downgrade_errors_for=(), use_closest_read_length=False):
        """
        Get the handler configurations for the specified parameters.

        :param instrument_and_reagent_type: type of instrument and reagents to match from config
        :param read_length: give the read length either as str or int
        :returns: the corresponding handler configuration(s)
        """

        try:
            handler_config = self._get_matching_handler(instrument_and_reagent_type,
                                                        read_length, use_closest_read_length)
            handler_config_with_defaults = self._add_default_config(handler_config)
            downgraded_handler_config_with_defaults = self._downgrade_errors(
                                                            handler_config_with_defaults,
                                                            downgrade_errors_for)
            return downgraded_handler_config_with_defaults
        except ConfigEntryMissing as e:
            log.error("Could not find a config entry for instrument '{}' "
                      "with read length '{}'. Please check the provided config "
                      "file ".format(instrument_and_reagent_type,
                                     read_length))
            raise e

    def __getitem__(self, key):
        return self._config[key]

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default
